/* Esse - Semplice-text-editor */

using GLib;
using Gtk;
using Glade;
using GConf;
using Pango;

public class Document : Gtk.ScrolledWindow
{
		// Attributi del documento
	public string filename = "";	// Nome del file
	public bool modified = false;	// flag che indica se il documento è stato modificato
	public Gtk.TextView textview;	// la view del documento, collegata alla tab corrispondente
	public Gtk.TextBuffer buffer;	// il buffer di testo del documento
	public Gtk.Label label;			// il titolo della tab
	
	construct {
			// Inizializza la view, imposta il font e abilita la visualizzazione
		this.textview = new Gtk.TextView();
		
		var font = GConf.Client.get_default().get_string("/desktop/gnome/interface/monospace_font_name");
		this.textview.modify_font(Pango.FontDescription.from_string(font));
		
		this.buffer = this.textview.get_buffer();
		this.add(this.textview);
		this.show_all();
	}

		// Carica un file dentro il tab. Se il file non esiste o non è leggibile, torna false
	public bool set_file(string filename)
	{
		string contents = "";
		
		if (filename != "" && !FileUtils.get_contents(filename, out contents)) {
			return false;
		}
		
		this.buffer.set_text(contents, -1);
		this.filename = filename;
		this.buffer.modified = 0;
		this.label.set_label(Path.get_basename(filename));
		this.label.set_tooltip_text(filename);
		return true;
	}

}

public class Gui : Gtk.Object {

	private const string program_name = "Esse";
	private const string version = "1.0";
	private const string copyright = "Alessandro Pellizzari <alex@amiran.it>";
	private const string license = "This software is licensed under the terms of GNU GPL2";
	
		// La variabile che conterrà il riferimento al file glade
	Glade.XML xml;
	
	construct
	{
			// Crea l'interfaccia dal file glade e connette automaticamente i segnali
			// corrispondenti ai widget
		this.xml = new Glade.XML("/usr/share/esse/esse.glade", null, null);
		this.xml.signal_autoconnect_full(this.connect_signals);
	}

	private void alert(string title="", string msg="")
	{
		var req = new Gtk.Dialog.with_buttons(title,
			(Gtk.Window) this,
			Gtk.DialogFlags.MODAL,
			Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT,
			null);
		
		var label = new Gtk.Label(msg);
		label.show();
		req.vbox.pack_start(label, true, true, 10);
		
		int response = req.run();
		req.close();
	}
	
	public void add_notebook_page(string filename="")
	{
			// Crea una nuova tab, ci associa una label standard per il titolo,
			// con il pulsante di chiusura, abbina le callback e abilita la
			// visualizzazione.
			// Se viene passato un nome di file, lo apre e lo inserisce nella tab
		if (filename != "" && !FileUtils.test(filename, GLib.FileTest.IS_REGULAR)) {
			string labeltxt = "Unable to open " + filename;
			this.alert("Unable to open file", labeltxt);
			return;
		}

		var doc = new Document();
		
		var buffer = (Gtk.TextBuffer) doc.buffer;
		buffer.set_data("document", doc);
		buffer.modified_changed += this.on_document_modified;
		
		var nb = (Gtk.Notebook) this.widget("nb_main");
		
		var hb = new Gtk.HBox(false, 0);
		var label = new Gtk.Label("New Document");
		doc.label = label;
		hb.pack_start( label , true, true, 0);
		
		var close = new Gtk.Button();
		close.relief = Gtk.ReliefStyle.NONE;
		close.set_image( new Gtk.Image.from_stock(Gtk.STOCK_CLOSE, Gtk.IconSize.MENU) );
		close.set_data("document", doc);
		
		close.clicked += this.on_tab_close;
		
		hb.pack_start( close , false, true, 0);
		hb.show_all();
		
		if (filename != "" && !doc.set_file(filename) ) {
			string labeltxt = "Unable to open " + filename;
			this.alert("Unable to open file", labeltxt);
		}
		
		nb.append_page( doc, hb);
		nb.set_current_page( NotebookPage.num(nb, doc) );
	}
	
	/**
	 * Funzione di appoggio che restituisce il widget con un certo nome
	 * prendendolo dalla struttura di glade
	 * 
	 */
	private Gtk.Widget widget(string name)
	{
		var w = this.xml.get_widget(name);
		return w;
	}
	
	/**
	 * Connette tutti i segnali ai widget, ispezionando automaticamente le
	 * funzioni definite per questa classe
	 */
	[InstanceLast]
	public void connect_signals(string handler_name, GLib.Object object, string signal_name, string? signal_data, GLib.Object? connect_object, bool after)
	{
		string handler_real_name = "gui_"+handler_name;
		
		GLib.Module module = GLib.Module.open(null, ModuleFlags.BIND_LAZY);
		
		void * sym;
		
		if(!module.symbol(handler_real_name, out sym)) {
			stdout.printf("Symbol not found: %s\n", handler_real_name);
		} else {
		 	GLib.Signal.connect(object, signal_name, (GLib.Callback) sym, this);
		}
	}

	[CCode (instance_pos = -1)]
	public bool on_quit(Gtk.Widget widget)
	{
		if (this.check_quit()) {
			Gtk.main_quit();
			return false;
		} else {
			return true;
		}
	}

	[CCode (instance_pos = -1)]
	public bool on_quit_event(Gtk.Object widget, Gdk.Event e)
	{
		if (this.check_quit()) {
			Gtk.main_quit();
			return false;
		} else {
			return true;
		}
	}


	private bool check_quit()
	{
		var nb = (Gtk.Notebook)this.widget("nb_main");
		int n_pages = nb.get_n_pages();
		for (int i=0; i<n_pages; i++) {
			Document doc = (Document)nb.get_nth_page(i);
			if ( doc.buffer.get_modified() ) {
				var req = new Gtk.Dialog.with_buttons("Document close",
					(Gtk.Window) this,
					Gtk.DialogFlags.MODAL,
					Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
					Gtk.STOCK_YES, Gtk.ResponseType.ACCEPT,
					null);
				
				var label = new Gtk.Label("There are some unsaved documents. Are you sure you want to quit?");
				label.show();
				req.vbox.pack_start(label, true, true, 10);
				
				int response = req.run();
				req.close();
				
				if (response == Gtk.ResponseType.ACCEPT) {
					return true;
				} else {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Se il documento viene modificato, cambia il titolo della tab per
	 * visualizzare un asterisco
	 */
	[CCode (instance_pos = -1)]
	public void on_document_modified(Gtk.TextBuffer widget)
	{
		if (widget.get_modified()) {
			var doc = (Document) (widget.get_data("document"));
			var nb = (Gtk.Notebook) this.widget("nb_main");
			string title = doc.label.get_label();
			
			if ( ! title.has_prefix("* ")) {
				title = "* "+title;
				doc.label.set_label(title);
			}
		}
	}
	
	/**
	 * Se una tab viene chiusa, controlla se il documento è stato modificato,
	 * e in caso presenta un requester per salvarlo o annullare la chiusura
	 */
	[CCode (instance_pos = -1)]
	public void on_tab_close(Gtk.Button widget)
	{
		var sw = (Document) widget.get_data("document");
		
		var nb = (Gtk.Notebook) this.widget("nb_main");
		int page = NotebookPage.num(nb,sw);
		
		if (sw.buffer.get_modified()) {
			
			// La imposta come pagina attiva
			nb.set_current_page(page);
			
			var req = new Gtk.Dialog.with_buttons("Document close",
						(Gtk.Window) this.widget("w_main"),
						Gtk.DialogFlags.MODAL,
						Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT,
						Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
						Gtk.STOCK_OK, Gtk.ResponseType.OK,
						null);
			
			var label = new Gtk.Label("Are you sure you want to close this unsaved document?");
			label.show();
			req.vbox.pack_start(label, true, true, 10);
			
			int response = req.run();
			req.close();
			
			if (response == (int)Gtk.ResponseType.OK) {
				// Chiudi tutto senza salvare
				nb.remove_page(page);
			} else if (response == (int) Gtk.ResponseType.ACCEPT) {
				// Salva e chiudi
				this.on_file_save( (Gtk.Widget) sw );
				if (sw.buffer.modified == 0) {
					nb.remove_page(page);
				}
			} else {
				// Non fare niente
			}
		} else {
			nb.remove_page(page);
		}
	}
	
	/**
	 * Aggiunge una tab
	 */
	[CCode (instance_pos = -1)]
	public void on_file_new(Gtk.Widget widget)
	{
		this.add_notebook_page();
	}

	/**
	 * Carica un file, prepara la tab per contenerlo e imposta le variabili
	 * di stato relative
	 */
	[CCode (instance_pos = -1)]
	public void on_file_open(Gtk.Widget widget)
	{
		var filereq = new Gtk.FileChooserDialog("Choose file", 
						(Gtk.Window) this.widget("w_main"), Gtk.FileChooserAction.OPEN,
						Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
						Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
						null);
		
		if (filereq.run() == Gtk.ResponseType.ACCEPT) {
			string filename = filereq.get_filename();
			filereq.close();

			var nb_main = (Gtk.Notebook) this.widget("nb_main");
			var active_doc = (Document) nb_main.get_nth_page(nb_main.get_current_page());

			if (active_doc.filename == "") {
				if ( !FileUtils.test(filename, GLib.FileTest.IS_REGULAR) || !active_doc.set_file(filename) ) {
					string labeltxt = "Unable to open " + filename;
					this.alert("Unable to open file", labeltxt);
				}
			} else {
				this.add_notebook_page(filename);
			}
		} else {
			filereq.close();
		}
	}

	/**
	 * Salva il file, eventualmente chiedendone il nome tramite requester,
	 * e imposta lo stato relativo del documento
	 */
	[CCode (instance_pos = -1)]
	public void on_file_save(Gtk.Widget widget)
	{
		bool save = true;
		var nb_main = (Gtk.Notebook) this.widget("nb_main");
		var active_doc = (Document) nb_main.get_nth_page(nb_main.get_current_page());
		
		if (!active_doc.buffer.get_modified()) {
			return;
		}
		
		if (active_doc.filename == "") {
			this.on_file_saveas(widget);
		} else {
			Gtk.TextIter s, e;
			active_doc.buffer.get_bounds(out s, out e);
			string contents = active_doc.buffer.get_text(s, e, false);
			if (!FileUtils.set_contents(active_doc.filename, contents, -1)) {
				this.alert("Errore", "Errore durante il salvataggio\n");
			}
			active_doc.buffer.modified = 0;
			active_doc.label.set_label(Path.get_basename(active_doc.filename));
			active_doc.label.set_tooltip_text(active_doc.filename);
		}
	}

	/**
	 * Presenta un requester che chiede il nome del file da salvare, imposta
	 * lo stato relativo e quindi richiama la funzione per il salvataggio
	 */
	[CCode (instance_pos = -1)]
	public void on_file_saveas(Gtk.Widget widget)
	{
		string filename = this.get_filename();
		if (filename != null) {
			var nb_main = (Gtk.Notebook) this.widget("nb_main");
			var active_doc = (Document) nb_main.get_nth_page(nb_main.get_current_page());
			active_doc.filename = filename;
			this.on_file_save(widget);
		}
	}
	
	/**
	 * Funzione di appoggio che apre un requester per chiedere il nome di un file
	 */
	public string get_filename()
	{
		string filename = null;
		var filereq = new Gtk.FileChooserDialog("Choose file name", 
						(Gtk.Window) this.widget("w_main"), Gtk.FileChooserAction.SAVE,
						Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
						Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
						null);
		
		if (filereq.run() == Gtk.ResponseType.ACCEPT) {
			filename = filereq.get_filename();
		}
		
		filereq.close();
		return filename;
	}		
	
	/**
	 * Taglia il testo selezionato dal buffer e lo salva in una clipboard
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_cut(Gtk.Widget widget)
	{
		var nb_main = (Gtk.Notebook) this.widget("nb_main");
		var active_doc = (Document) nb_main.get_nth_page(nb_main.get_current_page());
		var clipboard = active_doc.textview.get_clipboard(Gdk.SELECTION_CLIPBOARD);
		
		active_doc.buffer.cut_clipboard(clipboard, active_doc.textview.get_editable());
	}

	/**
	 * Copia il testo selezionato nel buffer in una clipboard
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_copy(Gtk.Widget widget)
	{
		var nb_main = (Gtk.Notebook) this.widget("nb_main");
		var active_doc = (Document) nb_main.get_nth_page(nb_main.get_current_page());
		var clipboard = active_doc.textview.get_clipboard(Gdk.SELECTION_CLIPBOARD);
		
		active_doc.buffer.copy_clipboard(clipboard);
	}

	/**
	 * Incolla il testo nella clipboard nel buffer alla posizione corrente
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_paste(Gtk.Widget widget)
	{
		var nb_main = (Gtk.Notebook) this.widget("nb_main");
		var active_doc = (Document) nb_main.get_nth_page(nb_main.get_current_page());
		var clipboard = active_doc.textview.get_clipboard(Gdk.SELECTION_CLIPBOARD);
		
		active_doc.buffer.paste_clipboard(clipboard, null, active_doc.textview.get_editable());

	}

	/**
	 * Cancella dal buffer il testo selezionato
	 */
	[CCode (instance_pos = -1)]
	public void on_edit_delete(Gtk.Widget widget)
	{
		var nb_main = (Gtk.Notebook) this.widget("nb_main");
		var active_doc = (Document) nb_main.get_nth_page(nb_main.get_current_page());
		
		active_doc.buffer.delete_selection(true, active_doc.textview.get_editable());
	}

	/**
	 * Presenta un about dialog con le informazioni sul software
	 */
	[CCode (instance_pos = -1)]
	public void on_help_info(Gtk.Widget widget)
	{
		var about = new Gtk.AboutDialog();
		about.program_name = this.program_name;
		about.version = this.version;
		about.copyright = this.copyright;
		about.license = this.license;
		about.run();
		about.close();
	}

}

/**
 * Classe principale del programma, il punto di entrata
 * Inizializza la GUI e fa girare il gestore di eventi di Gtk
 */
public class Esse : GLib.Object
{

	public static int main(string[] args)
	{
		Gtk.init(ref args);
		
		var gui = new Gui();

		int i=0;
		foreach (string arg in args) {
			if (i++ != 0) {
				gui.add_notebook_page(arg);
			}
		}
		if (i==1) {	
			// Nessun file sulla command line, aggiunge un tab vuoto
			gui.add_notebook_page();
		}
		
		Gtk.main();
		return 0;
	}
}
